#include <cstdio>
#include <cmath>

#include "core.hh"

#ifdef __sun
#ifdef __SVR4
#include <sys/audioio.h>
#else
#include <sun/audioio.h>
#endif
#endif

#ifdef linux
#include <sys/soundcard.h>
#endif

#ifdef __CYGWIN__
#include <windows.h>
#include <mmsystem.h>
#endif

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <string>

#include <termio.h>

using namespace std;

/* LOGMUL = exp(log(2) / 3) */
#define LOGMUL 1.259921049894873

static struct termio back;

void Exit ()
{
    if(ioctl(0, TCSETA, &back) < 0)
        fcntl(0, F_SETFL, fcntl(0, F_GETFL) &~ O_NONBLOCK);

    SPCCore_MuteSound();
    exit(0);
}

#include "argh.hh"

#define VERSIONSTR \
    "spcplay - SPC player based on OpenSPC\n" \
    "player copyright (C) 2001,2003 Bisqwit (http://iki.fi/bisqwit/)\n"

static string filee;
#ifdef linux
static string snd_dev = "/dev/dsp";
#endif
#ifdef __sun
static string snd_dev = "/dev/audio";
#endif
#ifdef __CYGWIN__
static string snd_dev = "(windows)";
namespace WindowsAudio
{
  const unsigned NUM_BUFS = 4;
  const unsigned BUF_SIZE = 4*32768;
  static HWAVEOUT hWaveOut;
  static WAVEHDR  bufs[NUM_BUFS];
  static unsigned head; // buffer currently playing
  static unsigned tail; // buffer currently ready

  static void Open()
  {
      WAVEFORMATEX wformat;
      MMRESULT result;
      
      const int channels = SPCCore_GetChannels();
      const int rate     = SPCCore_GetRate();
      const int bits     = 16;

      //fill waveformatex
      memset(&wformat, 0, sizeof(wformat));
      wformat.nChannels       = channels;
      wformat.nSamplesPerSec  = rate;
      wformat.wFormatTag      = WAVE_FORMAT_PCM;
      wformat.wBitsPerSample  = bits;
      wformat.nBlockAlign     = wformat.nChannels * (wformat.wBitsPerSample >> 3);
      wformat.nAvgBytesPerSec = wformat.nSamplesPerSec * wformat.nBlockAlign;
       
      //open sound device
      //WAVE_MAPPER always points to the default wave device on the system
      result = waveOutOpen
      (
        &hWaveOut,WAVE_MAPPER,&wformat,
        0,0,CALLBACK_NULL|WAVE_FORMAT_DIRECT
      );
      if(result == WAVERR_BADFORMAT)
      {
          fprintf(stderr, "ao_win32: format not supported\n");
          return;
      }
      if(result != MMSYSERR_NOERROR)
      {
          fprintf(stderr, "ao_win32: unable to open wave mapper device\n");
          return;
      }
      head=0;
      tail=0;
  }
  static void Close()
  {
      waveOutReset(hWaveOut);
      waveOutClose(hWaveOut);
  }
  static void ReallyWrite(const unsigned char* Buf, unsigned bufsize)
  {
  Retry:
      /* Flush all possible buffers */
      while(head != tail
         && waveOutUnprepareHeader(hWaveOut, &bufs[head], sizeof(*bufs)) != WAVERR_STILLPLAYING)
      {
          CHAR* copy = bufs[head].lpData;
          delete[] copy;
          head = (head+1) % NUM_BUFS;
      }
      
      if( (tail+1)%NUM_BUFS == head)
      {
          Sleep(10);
          goto Retry;
      }
      
      
      WAVEHDR* Work = &bufs[tail];
      tail = (tail+1) % NUM_BUFS;
      
      int res;
      memset(Work, 0, sizeof(*Work));
      CHAR* copy = new CHAR[bufsize];
      memcpy(copy,Buf,bufsize);
      Work->lpData         = copy;
      Work->dwBufferLength = bufsize;
      Work->dwLoops = 0;
      res = waveOutPrepareHeader(hWaveOut, Work, sizeof(*Work));
      //fprintf(stderr, "prep(%d)", res);
      res = waveOutWrite(hWaveOut, Work, sizeof(*Work));
      //fprintf(stderr, "out(%d)", res);
  }
  static void Write(const unsigned char* Buf, unsigned bufsize)
  {
      const unsigned COMBINE_LEN = BUF_SIZE;
      static unsigned char Combine[COMBINE_LEN];
      static unsigned CombinePtr=0;
      while(bufsize > 0)
      {
          unsigned take = COMBINE_LEN-CombinePtr;
          if(take > bufsize) take=bufsize;
          memcpy(&Combine[CombinePtr], Buf, take);
          bufsize -= take;
          CombinePtr += take;
          if(CombinePtr >= COMBINE_LEN)
          {
              ReallyWrite(Combine, COMBINE_LEN);
              CombinePtr = 0;
          }
      }
  }
}
#endif
static int surround = 0;
static int reversestereo = 0;
static int trackno = 0;

class Handle : public arghandler
{
    bool Help;
private:
    string opt_h(const string &s)
    {
        Help = true;
        return s;
    }
    string opt_V(const string &)
    {
        printf(VERSIONSTR);
        exit(0);
    }
    string opt_2(const string &s) { SPCCore_SetChannels(2);                   return s; }
    string opt_m(const string &s) { SPCCore_SetChannels(1);                   return s; }
    string opt_b(const string &s) { SPCCore_SetBitness(8);                 return s; }
    string opt_6(const string &s) { SPCCore_SetBitness(16);                return s; }
    string opt_n(const string &s) { SPCCore_SetInterpolation(0);           return s; }
    string opt_i(const string &s) { SPCCore_SetInterpolation(1);           return s; }
    string opt_e(const string &s) { SPCCore_SetEcho(1);                    return s; }
    string opt_c(const string &s) { SPCCore_SetEcho(0);                    return s; }
    string opt_a(const string &s) { SPCCore_SetAlt(1);    return s; }
    string opt_u(const string &s) { surround = 1;                       return s; }
    string opt_s(const string &s) { reversestereo = 1;                  return s; }
    string opt_r(const string &s)
    {
        const char *q = s.c_str();
        const char *p = q;
        long rate = strtol(p, (char **)&p, 10);
        if(rate < 18 || rate > 256000)
        {
            argerror(s);
        }
        else
            SPCCore_SetRate(rate);
        return s.substr(p-q);
    }
    string opt_t(const string &s)
    {
        const char *q = s.c_str();
        const char *p = q;
        trackno = strtol(p, (char **)&p, 10)-1;
        if(trackno<0)trackno=0;
        return s.substr(p-q);
    }
    string opt_o(const string &s)
    {
        snd_dev = s;
        return "";
    }
public:
    Handle(const char *defopts, int argc, const char *const *argv)
    : arghandler(defopts, argc, argv),
        Help(false)
    {
        add("-2", "--stereo", "Specifies stereo sound. Redundant. Default.", &Handle::opt_2);
        add("-s", "--reverse","Reverse stereo.",                             &Handle::opt_s);
        add("-m", "--mono",   "Specifies mono sound.",                       &Handle::opt_m);
        add("-b", "--byte",   "8-bit mode. Normally 16-bit mode.",           &Handle::opt_b);
        add("-6", "--word",   "16-bit mode. Redundant.",                     &Handle::opt_6);
        add("-r", "--rate",   "Specifies playing rate. 44410 is default.",   &Handle::opt_r);
        add("-i", "--ipolate","Enables interpolating. Default.",             &Handle::opt_i);
        add("-n", "--npolate","Disables interpolating. (Toggle-key: 9/i)",   &Handle::opt_n);
        add("-t", "--track",  "Selects track.",                              &Handle::opt_t);
        add("-e", "--echo",   "Enables echo. Default.",                      &Handle::opt_e);
        add("-c", "--noecho", "Disables echo. (Toggle-key: e)",              &Handle::opt_c);
        add("-o", "--output", "Specify output device (default: /dev/dsp).",  &Handle::opt_o);
        add("-h", "--help",   "Help",                                        &Handle::opt_h);
        add("-V", "--version","Version information",                         &Handle::opt_V);
        add("-a", "--altsampledecode",
                               "Better sample decoding (Toggle-key: a)",         &Handle::opt_a);
        add("-u", "--surround","Flip one channel (stereo only) (Toggle-key: u)", &Handle::opt_u);
    }
    virtual void defarg(const string &arg)
    {
        if(arg.size())filee = arg;
    }
    void Usage()
    {
        fprintf(stderr, VERSIONSTR);
        fprintf(stderr,
            "Usage: %s [<options> ...] <file>\n"
            "\n"
            "Emulates and plays SPC files (snes soundtracks).\n"
            "\n"
            "Options:\n", a0.c_str());
        
        listoptions();
        
        exit(0);
    }
    virtual void parse()
    {
        arghandler::parse();
        if(Help || !filee.size())Usage();
    }
};

static double mastervolume = 1;
static double extrastereo  = 1;

static int cur_rate;

static double seconds_played;

static void State(void)
{
     fprintf(stderr, "\rRate %-4d", cur_rate);
     fprintf(stderr, "   Volume %-4.0f ", (mastervolume*100+0.5));
     if(extrastereo != 1.0)
     {
         fprintf(stderr, "   Stereo %-4.0f%% ", (extrastereo*100+0.5));
     }
     for(unsigned a=0; a<8; ++a)
         fputc(SPCCore_GetChannelStatus(a) ? (a+'1') : '_', stderr);
     
     double time = seconds_played;
     double s = std::fmod(time, 60.0); time -= s; time /= 60.0;
     double m = std::fmod(time, 60.0); time -= m; time /= 60.0;
     double h = std::fmod(time, 24.0); time -= h; time /= 24.0;
     double d = time;
     
     fprintf(stderr, "  %g:%g:%02g:%02.1f  ", d,h,m,s);
     
     fprintf(stderr, "%10s", "");
     fflush(stderr);
}

static void KeyInit()
{
    struct termio term = back;
    
    // Disable linebuffer and echoing
    term.c_lflag &= ~(ICANON|ECHO);
    // Disable signals generated by ^C, ^Z and ^] (int,stop,quit)
    //term.c_lflag &= ~(ISIG);
    
    // Disable special handling of ^S and ^Q
    //term.c_iflag &= ~(IXON);
    
    term.c_cc[VMIN] = 0; // 0=no block, 1=do block
    if(ioctl(0, TCSETA, &term) < 0)
        fcntl(0, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);
}

static void Resume(int)
{
    KeyInit();
}

int main(int argc, const char *const *argv)
{
    Handle parameters("", argc, argv);
    parameters.parse();
    
    ioctl(0, TCGETA, &back);
    signal(SIGCONT, Resume);
    KeyInit();
    
    unsigned bufsize = SPCCore_Init();
    unsigned char *Buf = (unsigned char *)malloc(bufsize);
    
#if 0
    for(unsigned fni=0; fni<files.size(); ++fni)
#else
    for(unsigned i=0; i<1; ++i)
#endif
    {
#if 0
        const char *fn = files[fni].c_str();
#else
        const char *fn = filee.c_str();
#endif
        if(SPCCore_Load(fn))
        {
            string s = fn;
            s += " could not be opened";
            perror(s.c_str());
            break;
        }
        
        int fd;
        
        if(snd_dev=="-" || !snd_dev.size())
            fd = STDOUT_FILENO;
        else
        {
#ifdef __sun
            fd = open(snd_dev.c_str(), O_WRONLY);
            if(fd >= 0)
            {
                audio_info_t audio;
                AUDIO_INITINFO(&audio);
                audio.play.sample_rate = SPCCore_GetRate();
                audio.play.channels = SPCCore_GetChannels();
                audio.play.precision = SPCCore_GetBitness();
                audio.play.encoding = AUDIO_ENCODING_LINEAR;
                ioctl(fd, AUDIO_SETINFO, &audio);
            }
#endif
#ifdef linux
            fd = open(snd_dev.c_str(), O_WRONLY);
            if(fd >= 0)
            {
                int stereo = SPCCore_GetChannels()-1;
                int res = SPCCore_GetBitness();
                int rate = SPCCore_GetRate();
                ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &res);
                ioctl(fd, SNDCTL_DSP_STEREO,     &stereo);
                ioctl(fd, SNDCTL_DSP_SPEED,      &rate);
                SPCCore_SetBitness(res);
                SPCCore_SetRate(rate);
                SPCCore_SetChannels(stereo?2:1);
            }
#endif
#ifdef __CYGWIN__
            fd = 0;
            WindowsAudio::Open();
#endif
        }
        
        if(fd < 0)
        {
            string s = snd_dev + " could not be opened";
            fprintf(stderr, "%s\n", s.c_str());
            break;
        }
        
        //memset(Buf, 0, bufsize);
        //write(fd, Buf, bufsize);
        
        fprintf(stderr, "File: %s\n", fn);
        
        SPCCore_SetPlaybackSpeed(cur_rate = 100);
        SPCCore_SelectTrack(trackno);
                
        SPCCore_ID *id = SPCCore_GetID(fn);
        if(id)
        {
            char Unkn[64];sprintf(Unkn,"Unknown (%d)", (int)id->emulator);
            fprintf(stderr,
                 "Song: %-32s[%d/%d]Game: %-32s\n"
                 "Comments: %-32s  Author: %-32s\n"
                 "Dumper: %-16s    Playtime: %d   Fadetime: %d   Emulator: %s\n",
                 id->songname,
                 trackno+1,
                 SPCCore_GetNumTracks(),
                 id->gametitle,
                 id->comments,
                 id->author,
                 id->dumper,
                 id->playtime,
                 id->fadetime,
                 (id->emulator==1) ? "ZSnes"
               : (id->emulator==2) ? "snes9x"
               : Unkn);
        }
        else
            fprintf(stderr, "(no id666 tag)\n");
        
        State();
        nice(-40);
        
        fcntl(1, F_SETFL, fcntl(1, F_GETFL) & ~O_NONBLOCK);
        for(;;)
        {
            bool ignore_write = false;
            unsigned char c;
            while(read(0, &c, 1) == 1)
            {
                if(c=='!') { ignore_write = true; break; }
                
                if(c=='q')goto Done;
                if(c=='u')surround ^= 1;
                if(c=='s')reversestereo ^= 1;
                if(c=='a')SPCCore_SetAlt(!SPCCore_GetAlt());
                if(c=='9' || c=='i')
                    SPCCore_SetInterpolation(
                      (SPCCore_GetInterpolation()+1) % SPCCore_GetNumInterpolation()
                                            );
                if(c=='e')SPCCore_SetEcho(!SPCCore_GetEcho());
                if(c=='c')
                {
                    if(cur_rate<50)++cur_rate;else cur_rate+=10;
                    SPCCore_SetPlaybackSpeed(cur_rate);
                    cur_rate = SPCCore_GetPlaybackSpeed();
                }
                if(c=='d')
                {
                    if(cur_rate>50)cur_rate-=10;else
                    if(cur_rate>1)--cur_rate;
                    SPCCore_SetPlaybackSpeed(cur_rate);
                    cur_rate = SPCCore_GetPlaybackSpeed();
                }
                
                if(c=='>' || c=='n')
                {
                  SPCCore_SelectTrack( (trackno=trackno+1) % SPCCore_GetNumTracks() );
                  seconds_played = 0;
                }
                if(c=='<' || c=='p')
                {
                  SPCCore_SelectTrack( (trackno=trackno+SPCCore_GetNumTracks()
                                                       -1) % SPCCore_GetNumTracks() );
                  seconds_played = 0;
                }
                
                if(c=='+'||c=='*')
                    mastervolume *= LOGMUL;
                if(c=='-'||c=='/')
                {
                    mastervolume /= LOGMUL;
                    if(mastervolume < 0.004)mastervolume = 0.004;
                }
                if(c=='w')
                    extrastereo *= LOGMUL;
                if(c=='W')
                {
                    extrastereo /= LOGMUL;
                    if(extrastereo < 0.00984)mastervolume = 0.00984;
                }
                if(c>='1' && c<='8')
                {
                    SPCCore_SetChannelStatus(c-'1',
                   !SPCCore_GetChannelStatus(c-'1'));
                }
                else if(c=='0')
                {
                    for(unsigned n=0; n<8; ++n)
                        SPCCore_SetChannelStatus(n, 1);
                }
                State();
            }
            
            SPCCore_Generate(Buf);
            
            if(ignore_write) goto IgnoreWrite;
            
            /* LITTLE-ENDIAN! */
            if(SPCCore_GetBitness() == 16)
            {
                signed short *smp = (signed short *)Buf;
                const unsigned count = bufsize / sizeof(*smp);
                for(unsigned a=0; a<count; ++a)
                {
                    int v = static_cast<int> (*smp * mastervolume);
                    if(surround && (a&1) && SPCCore_GetChannels() > 1)v=-v;
                    if(v > 32767)v = 32767;else if(v < -32768)v = -32768;
                    *smp++ = v;
                }
                if(reversestereo && SPCCore_GetChannels() == 2)
                {
                    smp = (signed short *)Buf;
                    for(unsigned a=0; a<count; ++a)
                    {
                        int v1 = smp[0];
                        int v2 = smp[1];
                        smp[0] = v2;
                        smp[1] = v1;
                        smp += 2;
                        ++a;
                    }
                }
                if(extrastereo != 1.0 && SPCCore_GetChannels() == 2)
                {
                    smp = (signed short *)Buf;
                    for(unsigned a=0; a<count; ++a)
                    {
                        int v1 = smp[0];
                        int v2 = smp[1];
                        
                        double avg = (v1 + v2) / 2.0;
                        int l = avg + extrastereo * (v1 - avg);
                        int r = avg + extrastereo * (v2 - avg);
                        if(l > 32767)l = 32767;else if(l < -32768)l = -32768;
                        if(r > 32767)r = 32767;else if(r < -32768)r = -32768;
                        
                        smp[0] = (int)l;
                        smp[1] = (int)r;
                        smp += 2;
                        ++a;
                    }
                }
            }
            else
            {
                /* FIXME: Volume is ignored for 8-bit */
            }
            
#ifdef __CYGWIN__
            if(fd >= 0)
                WindowsAudio::Write(Buf, bufsize);
            else
#endif
            write(fd, Buf, bufsize);
IgnoreWrite: ;
            seconds_played += (bufsize / (SPCCore_GetBitness() / 8)
                                      / (SPCCore_GetChannels())
                                      / (double)SPCCore_GetRate())
                                  / (cur_rate / 100.0);
            static double last_sec = 0;
            if((int)last_sec != (int)seconds_played)
            {
                State();
                last_sec = seconds_played;
            }
        }
Done:
#ifdef __CYGWIN__
        if(fd == 0)
            WindowsAudio::Close();
        else
#endif
        close(fd);
    }
    
    free(Buf);
    
    SPCCore_Terminate();
    
    Exit();
    return 0;
}
